/* icondm - Icon DE Display Manager
*/

#ifndef _APP_H_
#define _APP_H_

#include <X11/Xlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <setjmp.h>
#include <stdlib.h>
#include <iostream>
//#include "panel.h"
#include "cfg.h"
#include "image.h"
#include "button.h"

class App {
public:
    App(int argc, char** argv);
    ~App();
    void Run();
    int GetServerPID();
    void StopServer();

	bool serverStarted;
    // Lock functions
    void GetLock();
    void RemoveLock();

private:
    void Login();
    void Reboot();
    void Halt();
    void Exit();
    void KillAllClients(Bool top);
    void RestartServer();
    void ReadConfig();
    void OpenLog();
    void CloseLog();
    void CreateServerAuth();
    char* StrConcat(const char* str1, const char* str2);
    void UpdatePid();

    bool AuthenticateUser(int mode);
 
    static void replaceVariables(std::string& input,
                                 const std::string& var,
                                 const std::string& value);

    // Server functions
    int StartServer();
    int ServerTimeout(int timeout, char *string);
    int WaitForServer();

    // Private data
    Window Root;
    Display* Dpy;
    int Scr;
//    Panel* LoginPanel;
    imenu* m_menu;
    int ServerPID;
    const char* DisplayName;

    // Options
    char* DispName;

    Cfg *cfg;

    Pixmap BackgroundPixmap;

    void blankScreen();
    Image* image;
    void setBackground(const std::string& filename);

    bool firstlogin;
    bool daemonmode;
    bool force_nodaemon;
	// For testing themes
//	char* testtheme;
    bool testing;
    
//    std::string themeName;
    std::string mcookie;

    const int mcookiesize;
};


#endif

